from jinja2 import Environment, BaseLoader, PackageLoader, Template


def workflow(workflow, runtime, username):

    template = get_template('workflow.yaml')
    return template.render(workflow=workflow, \
                            runtime=runtime, \
                        ).format(username=username, nb_user=runtime['NB_USER'])


def cronworkflow(workflow_yaml, schedule):
    template = get_template('cronworkflow.yaml')
    return template.render(name=workflow_yaml['metadata']['generateName'], \
                            workflow=workflow_yaml['spec'], \
                            schedule=schedule)


def get_template(name):
    env = Environment(loader=PackageLoader('jupyterflow', 'templates'), extensions=['jinja2_ansible_filters.AnsibleCoreFiltersExtension'])
    return env.get_template(name)



# def wf_run():
#     wf_path = sys.argv[3]
#     rendered = render_wf(wf_path)
#     body = yaml.safe_load(rendered)
#     print(request_object('workflows', body))
    

# def render_wf(wf_path):
#     abs_wf_path = os.path.abspath(wf_path)
#     abs_wf_dir = os.path.dirname(abs_wf_path)

#     with open(abs_wf_path) as f:
#         conf = yaml.safe_load(f)

#     jobs = conf['jobs']
#     d = dict(zip(range(1,len(jobs)+1), jobs))
#     dags = conf.get('dags', [])
#     resolve_tree = {}

#     for k in d.keys():
#         resolve_tree['job-' + str(k)] = []

#     for dag in dags:
#         pre, pro = dag.replace(' ', '').split('>>')
#         resolve_tree['job-' + pro].append('job-' + pre)

#     user_name = os.environ['JUPYTERHUB_USER']
#     escaped_name = get_escaped_user(user_name)
    
#     wf_jobs = []
#     for i, j in enumerate(jobs):
#         job = {}
#         job['name'] = 'job-' + str(i+1)
#         cmds = j.split(' ')
#         job['command'] = cmds
#         try:
#             job['dependencies'] = resolve_tree['job-' + str(i+1)]
#         except:
#             job['dependencies'] = []
#         wf_jobs.append(job)

#     defaults = {
#         "LIMITS_CPU": "8",
#         "LIMITS_MEM": "20Gi"
#     }
#     if os.path.isfile(CONFIG_FILE):
#         with open(CONFIG_FILE) as f:
#             config_file = yaml.safe_load(f)
#             for k, v in config_file.items():
#                 defaults[k] = v

#     cpu = get_value('LIMITS_CPU', defaults)
#     mem = get_value('LIMITS_MEM', defaults)
#     runAsUser = os.getuid() 

#     wf = {}
#     wf['jobs'] = wf_jobs
#     wf['escaped_name'] = escaped_name
#     wf['username'] = user_name
#     wf['workingDir'] = abs_wf_dir
#     wf['PATH'] = os.environ['PATH']
#     wf['runAsUser'] = runAsUser
#     wf['limits'] = {
#       "cpu": cpu,
#       "memory": mem
#     }

#     template = Environment(loader=BaseLoader).from_string(WF_TEMPLATE)
#     return template.render(**wf)



import json
import yaml

from kubernetes import config
from kubernetes import client


def create_object(body, namespace):
    group, version = body['apiVersion'].split('/')
    plural = body['kind'].lower() + 's'

    config.load_incluster_config()
    api_instance = client.CustomObjectsApi()
    response = api_instance.create_namespaced_custom_object(group, version, namespace, plural, body)
    
    return response


def get_notebook_pod(escaped_username, namespace):
    config.load_incluster_config()
    api_instance = client.CoreV1Api()
    label_selector = 'jupyterflow/username=%s' % escaped_username
    response = api_instance.list_namespaced_pod(namespace, label_selector=label_selector, _preload_content=False)
    
    return json.loads(response.data)['items'][0]



import os

import click
from click import ClickException

from . import workflow
from . import printer
from . import utils
from .runtime import runtime


@click.group()
def main():
    pass


@main.command()
@click.option('-f', '--filename', help='Path for workflow.yaml file. ex) \'jupyterflow run -f workflow.yaml\'', default=None)
@click.option('-c', '--command', help="Command to run workflow. ex) \'jupyterflow run -c \"python main.py >> python next.py\"\'", default=None)
@click.option('-o', '--output', help='Output format. default is \'-o jsonpath="metadata.name"\'', default='jsonpath="metadata.name"')
@click.option('--dry-run', help='Only prints Argo Workflow object, without accually sending it', default=False, is_flag=True)
def run(filename, command, output, dry_run):

    if command is not None:
        user_workflow = workflow.load_from_command(command)
        workingDir = os.getcwd()
    elif filename is not None:
        if not os.path.isfile(filename):
            raise ClickException("No such file %s" % filename)
        user_workflow = workflow.load_from_file(filename)
        workingDir = os.path.dirname(os.path.abspath(filename))
    else:
        raise ClickException("Provide either `-f` or `-c` option")

    runtime['workingDir'] = workingDir
    namespace = runtime['namespace']
    conf = utils.load_config()
    wf = workflow.build(user_workflow, namespace, runtime, conf)
    
    if dry_run:
        response = wf
        output = 'yaml'
    else:
        response = workflow.run(wf, namespace)

    printer.format(response, output)


@main.command()
@click.option('--generate-config', help='Generate config file', default=False, is_flag=True)
def config(generate_config):
    if generate_config:
        utils.create_config()
    else:
        conf = utils.load_config()
        printer.format(conf, 'yaml')
        


# @main.command()
# # @click.option('--generate-config/--no', help='Generate config', default=False)
# def test():
#     escaped_username = utils.get_escaped_user(os.environ['JUPYTERHUB_USER'])
#     with open('/var/run/secrets/kubernetes.io/serviceaccount/namespace') as f:
#         ns = f.read()
#     pod = k8s_client.get_notebook_pod(escaped_username, ns)
#     pod = pod.to_dict()
#     pod['metadata']['name']
#     pod['metadata']['namespace']



import os
import pwd

from . import utils

class Runtime(dict):
    def __init__(self, *args, **kwargs):
        self.update(*args, **kwargs)

        with open('/var/run/secrets/kubernetes.io/serviceaccount/namespace') as f:
            dict.__setitem__(self, 'namespace', f.read())
        
        NB_USER = os.environ['NB_USER']
        nb_user_pwd = pwd.getpwnam(NB_USER)
        dict.__setitem__(self, 'image', os.environ['JUPYTER_IMAGE_SPEC'])
        dict.__setitem__(self, 'workingDir', None)
        dict.__setitem__(self, 'PATH', os.environ['PATH'])
        dict.__setitem__(self, 'HOME', os.environ['HOME'])
        dict.__setitem__(self, 'NB_USER', NB_USER)
        dict.__setitem__(self, 'runAsUser', nb_user_pwd.pw_uid)
        dict.__setitem__(self, 'runAsGroup', nb_user_pwd.pw_gid)
        dict.__setitem__(self, 'escaped_username', utils.get_escaped_user(os.environ['JUPYTERHUB_USER']))


runtime = Runtime()





import yaml

from . import k8s_client
from . import render
from .runtime import runtime


def load_from_command(c):
    jobs = list(map(lambda s: s.strip(), c.split('>>')))
    dags = []
    for i, j in enumerate(jobs):
        if i == len(jobs) -1:
            break
        dags.append(">>".join([str(i+1), str(i+2)]))
    user_workflow = {}
    user_workflow['jobs'] = jobs
    user_workflow['dags'] = dags
    return user_workflow


def load_from_file(f):
    with open(f) as yaml_file:
        return yaml.safe_load(yaml_file)


def build(wf, namespace, runtime, config):
    escaped_username = runtime['escaped_username']
    #########################
    # resolve tree
    #########################
    workflow = {}
    workflow['name'] = wf.get('name', escaped_username)
    workflow['jobs'] = []

    cmd_mode = wf.get('cmd_mode', 'exec')

    jobs = wf['jobs']
    d = dict(zip(range(1,len(jobs)+1), jobs))
    dags = wf.get('dags', [])
    resolve_tree = {}

    for k in d.keys():
        resolve_tree['job-' + str(k)] = []

    for dag in dags:
        pre, pro = dag.replace(' ', '').split('>>')
        resolve_tree['job-' + pro].append('job-' + pre)

    for i, j in enumerate(jobs):
        job = {}
        job['name'] = 'job-' + str(i+1)
        if cmd_mode == 'exec':
            cmds = j.split(' ')
        else:
            cmds = ["/bin/sh", "-c", j]
        job['command'] = cmds
        if 'job-' + str(i+1) in resolve_tree:
            job['dependencies'] = resolve_tree['job-' + str(i+1)]
        else:
            job['dependencies'] = []
        workflow['jobs'].append(job)
    
    pod = k8s_client.get_notebook_pod(escaped_username, namespace)
    workflow['spec'] = build_wf_spec_from(pod)
    override_wf(workflow, config)

    ###########################
    # render workflow
    ###########################
    rendered_wf = render.workflow(
            workflow=workflow, \
            runtime=runtime, \
            username=escaped_username
    )
    workflow_yaml = yaml.safe_load(rendered_wf)

    if 'schedule' in wf:
        rendered_wf = render.cronworkflow(workflow_yaml, wf['schedule'])
        workflow_yaml = yaml.safe_load(rendered_wf)
    return workflow_yaml
        


def run(wf, namespace):
    response = k8s_client.create_object(wf, namespace)
    return response
    

def build_wf_spec_from(pod):
    spec = {}

    spec['serviceAccountName'] = pod['spec'].get('serviceAccountName', None)
    spec['image'] = runtime['image']
    spec['imagePullPolicy'] = pod['spec']['containers'][0].get('imagePullPolicy', 'Always')
    spec['imagePullSecrets'] = pod['spec'].get('imagePullSecrets', None)
    spec['runAsUser'] = runtime['runAsUser']
    spec['runAsGroup'] = runtime['runAsGroup']
    spec['env'] = pod['spec']['containers'][0].get('env', [])
    spec['env_from'] = pod['spec']['containers'][0].get('env_from', None)
    spec['resources'] = pod['spec']['containers'][0].get('resources', None)
    spec['volumeMounts'] = pod['spec']['containers'][0].get('volumeMounts', None)
    spec['volumes'] = pod['spec'].get('volumes', None)
    spec['nodeSelector'] = pod['spec'].get('nodeSelector', None)

    return spec


def override_wf(workflow, config):
    spec = config.get('spec', {})
    wf_spec = workflow['spec']
    
    wf_env = wf_spec['env']
    wf_vol = wf_spec['volumes']
    wf_volMount = wf_spec['volumeMounts']

    del wf_spec['env']
    del wf_spec['volumes']
    del wf_spec['volumeMounts']

    wf_spec.update(spec)

    if 'env' not in wf_spec:
        wf_spec['env'] = []
    wf_spec['env'].extend(wf_env)

    if 'volumes' not in wf_spec:
        wf_spec['volumes'] = []
    wf_spec['volumes'].extend(wf_vol)

    if 'volumeMounts' not in wf_spec:
        wf_spec['volumeMounts'] = []
    wf_spec['volumeMounts'].extend(wf_volMount)



import os

import yaml
import escapism
import string
import pkgutil


def get_escaped_user(user):
    safe_chars = set(string.ascii_lowercase + string.digits)
    safe_servername = escapism.escape(user, safe=safe_chars, escape_char='-').lower()
    legacy_escaped_username = ''.join([s if s in safe_chars else '-' for s in user.lower()])
    return escapism.escape(user, safe=safe_chars, escape_char='-').lower()


def create_config():
    data = pkgutil.get_data(__name__, "templates/jupyterflow.yaml")

    home = os.environ['HOME']
    CONFIG_FILE = os.path.join(home, '.jupyterflow.yaml')
    with open(CONFIG_FILE, 'wt') as f:
        f.write(data.decode('utf-8'))


def load_config():
    home = os.environ['HOME']
    CONFIG_FILE = os.path.join(home, '.jupyterflow.yaml')
    
    if os.path.isfile(CONFIG_FILE):
        with open(CONFIG_FILE) as f:
            conf = yaml.safe_load(f)
            if conf is None:
                return {}
            return conf
    return {}



def hanle_exception():
    pass



# import pkgutil
# data = pkgutil.get_data(__name__, "templates/workflow.yaml")


import os, re
from setuptools import setup, find_packages

# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


ROOT = os.path.dirname(__file__)
VERSION_RE = re.compile(r'''__version__ = ['"]([0-9.]+)['"]''')


requires = [
    'escapism>=1.0',
    'PyYAML>=5.3',
    'kubernetes>=12.0',
    'Jinja2>=2.11',
    'jinja2-ansible-filters>=1.3',
    'click>=7.1',
    'jsonpath-ng>=1.5'
]

def get_version():
    init = open(os.path.join(ROOT, 'jupyterflow', '__init__.py')).read()
    return VERSION_RE.search(init).group(1)


setup(
    name="jupyterflow",
    version=get_version(),
    author="hongkunyoo",
    author_email="hongkunyoo@gmail.com",
    description="Run your workflow on JupyterHub",
    long_description=read('README.md'),
    long_description_content_type="text/markdown",
    url='https://github.com/hongkunyoo/jupyterflow',
    license="BSD 3-Clause",
    keywords="ctl, jupyterhub, pipeline, ML",
    packages=find_packages(),
    # package_data={'jupyterflow': ['templates/*']},
    install_requires=requires,
    include_package_data=True,
    scripts=['bin/jupyterflow']
)



import pprint
import yaml
from jsonpath_ng import jsonpath, parse


def format(response, output):
    if output.startswith('jsonpath'):
        output = output.split('=')[-1].strip('"')
        jsonpath_expr = parse(output)
        for i in [match.value for match in jsonpath_expr.find(response)]:
            print(i)
    elif output.startswith('yaml'):
        print(yaml.dump(response, default_flow_style=False))
    else:
        pprint.pprint(response)